/*
 * Copyright 2017 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.example.android.camera2basic;

import android.os.Bundle;
import android.support.v7.app.AppCompatActivity;

//later for trigger
import android.content.Context;

import android.content.Intent;
import android.os.Handler;
import android.os.Message;
import android.content.IntentFilter;

import android.view.KeyEvent;
import android.content.BroadcastReceiver;


import android.util.Log;

public class CameraActivity extends AppCompatActivity {

    private static final String TAG = "CameraActivity";

    private static final String ACTION_TV_INPUT_VGA_1 = "android.intent.action.TV_INPUT_VGA_1";

    private static final int MSG_CLOSE_APP = 1;

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_camera);
        if (null == savedInstanceState) {
            getSupportFragmentManager().beginTransaction()
                    .replace(R.id.container, Camera2BasicFragment.newInstance())
                    .commit();
        }


        IntentFilter mIntentFilter = new IntentFilter();
        mIntentFilter.addAction(ACTION_TV_INPUT_VGA_1);
        registerReceiver(trigerReceiver, mIntentFilter);

    }

    public boolean onKeyUp(int keyCode, KeyEvent event) {
        Log.d(TAG,"On key up called "+keyCode);
        if( keyCode == KeyEvent.KEYCODE_TV_INPUT_VGA_1){
            finish();
            return true;
        }
        return super.onKeyUp(keyCode,event);
    }


    private BroadcastReceiver trigerReceiver = new BroadcastReceiver() {
        @Override
        public void onReceive(Context context, Intent intent) {
            if(intent.getAction().equals(ACTION_TV_INPUT_VGA_1)){
                boolean level = (boolean)intent.getExtras().getBoolean("keydown");
                if(!level) {
                    Message msg = new Message();
                    msg.what = MSG_CLOSE_APP;
                    myHandler.sendMessage(msg);
                }
            }

        }
    };

    private Handler myHandler = new Handler() {
        public void handleMessage(Message msg) {
            switch (msg.what){
                case MSG_CLOSE_APP:
                    finish();
                    break;
            }
        }
    };


    public void onPause(){
        super.onPause();
    }

    public void onResume(){
        super.onResume();
    }
    public void onDestroy()
    {
        super.onDestroy();
        unregisterReceiver(trigerReceiver);
    }


}
